function [E, S] = ellIntersectionIAProjET(EE,Proj)
%
% [E, S] = ellIntersectionIAProjET(EE,Proj)
%
% Computes the ellipsoid in the intersection of two or more ellipsoids 
% such that the volume of it projection is maximized.
% 
% Proj must be the non-symmetric matrices such that
% Proj'*Proj = P is a projection matrix, i.e. P^2=P.
%
% Adapted from ELLIPSOID/ELLINTERSECTION_IA by Alex Kurzhanskiy.
% Shahab Kaynama, Oct 2014, kaynama@ece.ubc.ca

  global ellOptions;

  if ~isstruct(ellOptions)
    evalin('base', 'ellipsoids_init;');
  end

  dims = dimension(EE);
  mn   = min(min(dims));
  mx   = max(max(dims));

  if mn ~= mx
    error('ELLINTERSECTION_IA: all ellipsoids must be of the same dimension.');
  end

  [m, n] = size(EE);
  M      = m * n;
  EE     = reshape(EE, 1, M);
  zz     = zeros(mn, 1);
  I      = eye(mn);

  if ellOptions.verbose > 0
    fprintf('Invoking YALMIP...\n');
  end

  B      = sdpvar(mn, mn);
  d      = sdpvar(mn, 1);
  ll     = sdpvar(M, 1);

  
  cnstr = lmi;

  for i = 1:M
    [q, Q] = double(EE(i));
    
    if rank(Q) < mn
      Q = regularizeET(Q); %Modified
    end
    
    A     = ell_inv(Q);
    b     = -A * q;
    c     = q' * A * q - 1;
    X     = [(-ll(i,1)-c+b'*Q*b) zz'       (d+Q*b)'
             zz                  ll(i,1)*I B
             (d+Q*b)             B         Q]; %#ok<NASGU>

    cnstr = cnstr + set('X>=0');
    cnstr = cnstr + set('ll(i, 1)>=0');
  end


  S = solvesdp(cnstr, -logdet(Proj*B*Proj'), ellOptions.sdpsettings); %Modified.
  B = double(B);
  d = double(d);
  
  if rank(B) < mn
    B = regularizeET(B); %Modified
  end

  P = B * B';
  P = (1-ellOptions.abs_tol)*0.5*(P + P');

  E = ellipsoid(d, P);

  if nargout < 2
    clear S;
  end

  return;
