function [ g, kernels ] = tightViabilityKernel(max_steps, delta_fixed, delta_jitter, accuracy, num_input_samples, state_dx)
% tightViabilityKernel: demonstrate tight sampled data viability kernel algorithm.
%
%   [ g, kernels ] = tightViabilityKernel(steps, delta, accuracy, input_samples, grid_size)
%  
% A modified version of the viability kernel (or, more generally, the
% discriminating kernel) algorithm is tight in the sampled data
% environment, and this routine demonstrates that tightness by repeating
% the example from the NAHS paper.  In this particular case, the constraint
% set consists of a narrow vertical channel below the horizontal axis, and
% narrow diagonal channels above the horizontal axis. Unlike the NAHS
% algorithm, this version will show the small gaps in the upper channels
% (provided that the jitter is small enough).
%
%     \dot x_1 =  u
%     \dot x_2 = -1
%        with input |u| \leq +1.
%
% There are many options in the function which can only be modifed by
% changing the file.
%
% Input Parameters:
%
%   max_steps: Positive integer.  Optional.  Default = 3.
%
%   delta_fixed: Non-negative double.  Size of the guaranteed inter-sample
%   period.  Optional.  Default = 2.  Note delta_fixed + delta_jitter must
%   be positive (otherwise it is not a sampled data system).
%
%   delta_jitter: Non-negative double.  Amount of jitter in the sample
%   period.  Optional.  Default = 0.
%
%   accuracy: String.  Order of accuracy of the spatial and temporal
%   approximation schemes.  Choose between 'medium' (ENO2, RK2) or
%   'veryHigh' (WENO5, RK3).  The latter shows a definite improvement in
%   the sharpness of the corners of the reach tube, but requires about four
%   times as long.  Optional.  Default = 'veryHigh'.
%
%   num_input_samples: Positive integer.  Number of different values of the
%   input to test.  Equivalent to the size of the grid in the input
%   dimension (although such a grid is never constructed). Optional.
%   Default = 11.
%
%   state_dx: Positive double.  Grid cell size.  Dividing this by two will
%   multiply run time by 8; consequently, it is recommended to use higher
%   accuracy before reducing the grid size.  Optional.  Default = 0.05.
%
% Output Parameters:
%
%   g: grid structure.
%
%   kernels: Cell vector.  Element i holds the implicit surface function
%   for the sampled data viability kernel after (i-1) steps (so the initial
%   conditions can be found in entry 1).

% Copyright 2011 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Created by Ian Mitchell, 2009/11/28
% Modified to the viability kernel example 2012/09/28
% Modified to improved algorithm with jitter 2014/10/12

  %---------------------------------------------------------------------------
  % Tell m-lint not to flag unreachable code in this file: There are a
  % bunch of problem / solver parameters which are set at the beginning of
  % this file.  For any given value of these parameters, some code may be
  % unreachable, but we want to keep it because we may change these
  % parameters later.
  %#ok<*UNRCH>

  start_time_overall = cputime;

  %---------------------------------------------------------------------------
  %% Make sure we can see the kernel m-files.
  run('../addPathToKernel');

  %---------------------------------------------------------------------------
  %% Optional input arguments.
  if(nargin < 1)
    max_steps = 3;
  end
  if(nargin < 2)
    delta_fixed = 2;
  end
  if(nargin < 3)
      delta_jitter = 0;
  end
  if(nargin < 4)
    accuracy = 'veryHigh';
  end
  if(nargin < 5)
    num_input_samples = 11;
  end
  if(nargin < 6)
    state_dx = 0.05;
  end

  if(delta_fixed + delta_jitter <= 0)
    error('delta_fixed + delta_jitter must be positive but is %f', delta_fixed + delta_jitter);
  end
  
  %---------------------------------------------------------------------------
  %% PDE solver parameters.

  % Plot after each timestep?  Overrides t_plot.
  single_step = false;

  % Do some reinitialization after each delta timestep?
  perform_reinit = true;
  
  % Pause after each plot?
  pause_after_plot = false;

  % Delete previous plot before showing next?  Only applies to 2D
  % visualization; the 3D visualization is always cleared after every step.
  delete_last_plot = true;

  % Parameters for visualizeLevelSets.
  level = 0;
  % One visualization is done in 2D, one in 3D.
  display_type = 'contour';

  % What order of accuracy to use for the computation's spatial derivatives
  % and temporal integration?
  switch(accuracy)
    case 'veryHigh'
      derivFunc = @upwindFirstWENO5;
      integrator_func = @odeCFL3;
    case 'medium'
      derivFunc = @upwindFirstENO2;
      integrator_func = @odeCFL2;
    otherwise
      error('Unknown accuracy: %s', accuracy);
  end

  % Some integration parameters apply to any integrator.
  integrator_options = odeCFLset('factorCFL', 0.95, 'stats', 'on');
  if(single_step)
    integrator_options = odeCFLset(integrator_options, 'singleStep', 'on');
  end

  %---------------------------------------------------------------------------
  %% Optional and problem specific parameters.

  % Maximum absolute value of the input.
  input_bound = +1.0;

  % The computational domain used in the state space will stretch
  % this distance horizontally in both positive and negative directions,
  % and this distance vertically above zero (it will always extend
  % vertically below zero to -1).
  state_bound = 5.0;
  
  % How wide is the vertical channel?  The diagonal channels will be ~70%
  % of this width.
  half_channel_width = 0.25;

  % Use reach tube (invariance kernel) or reach set for sample times 2...N?
  % Using the reach tube is equivalent to the NAHS algorithm, whereas using
  % the reach set is the new algorithm.
  use_reach_set = true;

  %---------------------------------------------------------------------------
  %% Create the grid.
  % We run the level set algorithm only in the state dimensions and make a
  % separate run for each sample of the input.  We can do that because the
  % dynamics in the input dimension is zero.  It also turns out to be more
  % accurate and faster this way, because there is no numerical dissipation
  % or artificial boundary conditions in the input dimension and we can use
  % a lower dimensional grid (less memory traffic).
  g.dim = 2;
  g.min = [ -state_bound;         -1.0 ];
  g.max = [ +state_bound; +state_bound ];
  g.dx = state_dx;
  g.bdry = @addGhostExtrapolate;
  g = processGrid(g);
  
  %---------------------------------------------------------------------------
  %% Create initial conditions.

  % Create the vertical channel below the horizontal axis.  Note that we
  % extend this vertical channel above the axis by the half_channel_width.
  % If we stop exactly at the axis, then the final implicit surface
  % function (constructed by lots of unions) gets very close to zero at the
  % axis and that can cause numerical issues later on.
  data_vertical_channel = shapeRectangleByCorners(g, [ -half_channel_width; -inf ], [ +half_channel_width; +half_channel_width ]);

  % Create the left and right diagonal channels.
  data_lower_left  = shapeHyperplane(g, [ -1; -1 ], [ -half_channel_width; 0.0 ]);
  data_upper_left  = shapeHyperplane(g, [ +1; +1 ], [ 0.0; +half_channel_width ]);
  data_lower_right = shapeHyperplane(g, [ +1; -1 ], [ +half_channel_width; 0.0 ]);
  data_upper_right = shapeHyperplane(g, [ -1; +1 ], [ 0.0; +half_channel_width ]);
  data_upper = shapeHyperplane(g, [ 0; -1 ], [ 0.0; 0.0 ]);
  data_left_channel = shapeIntersection(shapeIntersection(data_lower_left, data_upper_left), data_upper);  
  data_right_channel = shapeIntersection(shapeIntersection(data_lower_right, data_upper_right), data_upper);

  % Put it all together.
  data = shapeUnion(shapeUnion(data_left_channel, data_right_channel), data_vertical_channel);

  % Store the 2D initial conditions.
  kernels = cell(max_steps + 1, 1);
  kernels{1} = data;
  
  %---------------------------------------------------------------------------
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %---------------------------------------------------------------------------
  %% Set up PDE terms that implement the dynamics (reach set).
  % These will be set up to do the reach set calculation (eg: the implicit
  % surface is not restricted to only grow).

  % Since there are no disturbances, we can compute the purely convectional
  % velocity field.
  scheme_func_set = @termConvection;
  scheme_data_set.grid = g;
  scheme_data_set.derivFunc = derivFunc;

  % We will loop over input samples to approximate the invariance kernel.
  % We build the corresponding velocity fields here, and then we will plug
  % them in to scheme_data below.
  input_samples = linspace(-input_bound, +input_bound, num_input_samples);
  velocity_samples = cell(num_input_samples, 1);
  for i_input_samples = 1 : num_input_samples
    velocity_samples{i_input_samples} = { -input_samples(i_input_samples); +1 };
  end

  %---------------------------------------------------------------------------
  %% Set up a version of the PDE terms for the reach tube.
  % These are the same as the reach set terms, except that we restrict the
  % Hamiltonian so that reachable set only grows.  Whatever scheme handles
  % the reach set dynamics must already be completely set up.

  % Wrap the true Hamiltonian inside the term approximation restriction routine.
  scheme_func_tube = @termRestrictUpdate;
  scheme_data_tube.innerFunc = scheme_func_set;
  scheme_data_tube.innerData = scheme_data_set;
  % Zero sublevel set can only shrink.
  scheme_data_tube.positive = true;
  
  %---------------------------------------------------------------------------
  %% Set up reinitialization (for both set and tube) if necessary.
  if perform_reinit
    % We won't do reinitialization as part of the convection step, but
    % rather run several iterations at the end of each delta timestep.
    % Therefore, we don't need to register a post-timestep routine with the
    % integrator; instead, we'll call it directly.
    scheme_data_set.reinitAccuracy = accuracy;
    scheme_data_tube.reinitAccuracy = accuracy;
    % Do enough reinitialization steps to stretch across the channel.
    scheme_data_set.reinitSteps = round(half_channel_width / state_dx);
    scheme_data_tube.reinitSteps = round(half_channel_width / state_dx);
    % Note that the post-timestep routine is called with the outer
    % schemeData for the tube calculation, so we need to ensure that the
    % grid is available for the tube.
    scheme_data_tube.grid = g;
  end

  %---------------------------------------------------------------------------
  %% Initialize Display
  
  f = figure; %#ok<NASGU>
  h = visualizeLevelSet(g, kernels{1}, display_type, level, 'constraint set K_0');
  set(h, 'LineStyle', '-', 'LineWidth', 1, 'LineColor', 'red');
  axis(g.axis);
  daspect([ 1 1 1 ]);
  xlabel('x_1');  ylabel('x_2');
  hold on;
  
  drawnow;
  
  fprintf('Initialization time %g\n', cputime - start_time_overall);

  %---------------------------------------------------------------------------
  %% Prepare for the main loops.
  step_now = 1;
  
  fprintf('\nStarting computation of viability kernel.\n');
  start_time_kernel = cputime;
  
  %---------------------------------------------------------------------------
  %% Loop over time samples.
  while(step_now <= max_steps)

    % Initial conditions for this time step (reshaped into the
    % form appropriate for the ODE solver call).
    y_init = data(:);
    y_end = [];

    % How far to step?
    if((step_now > 1) && use_reach_set)
      % In the new version of the algorithm, on the second and subsequent
      % steps we use a reach set for delta_fixed and then a reach tube /
      % invariance kernel for delta_jitter.
      if(delta_fixed > 0)
        integrate_reach_set = true;
        t_span_set = [ 0, delta_fixed ];
      else
        integrate_reach_set = false;
        t_span_set = nan;
      end
      if(delta_jitter > 0)
        integrate_reach_tube = true;
        t_span_tube = [ 0, delta_jitter ];
      else
        integrate_reach_tube = false;
        t_span_tube = nan;
      end
    else
      % If this is the first sample period or we are using the old version
      % fo the algorithm, then use a reach tube / invariance kernel for the
      % whole period.
      integrate_reach_set = false;
      t_span_set = nan;
      integrate_reach_tube = true;
      t_span_tube = [ 0, delta_fixed + delta_jitter ];
      % We will also need a variable to store the invariance kernel from
      % this first step for each input so that we can perform intersections
      % on subsequent steps.
      first_invariant = cell(num_input_samples, 1);
    end

    start_time_sample = cputime;

   %-----------------------------------------------------------------------
    %% Loop over input samples
    for i_input_samples = 1 : num_input_samples

      % Record the flow field for this input sample in both the reach set
      % and reach tube schemeData structures.
      scheme_data_set.velocity = velocity_samples{i_input_samples};
      scheme_data_tube.innerData.velocity = velocity_samples{i_input_samples};

      % Take a timestep. Note that the different input samples may have
      % slightly different end times, and we are not properly adjusting
      % for that fact when we ignore the first return parameter
      
      % First the reach set (fixed sample time) if any.
      if integrate_reach_set
        [ ~, y_set ] = feval(integrator_func, scheme_func_set, t_span_set, y_init, integrator_options, scheme_data_set);
      else
        y_set = y_init;
      end
      % Then the reach tube if any.
      if integrate_reach_tube
        [ ~, y_tube ] = feval(integrator_func, scheme_func_tube, t_span_tube, y_set, integrator_options, scheme_data_tube);
      else
        y_tube = y_set;
      end

      if(step_now == 1)
        % Record the invariance kernel for future steps.
        first_invariant{i_input_samples} = y_tube;
        y_intersect = y_tube;
      else
        % Intersect the invariance kernel result with the one from the
        % first timestep.  Note that this call may generate warnings about
        % empty implicit surface functions for some input values.
        % Unfortunately, stupid Matlab does not provide an easy way to
        % catch warnings (whereas you can catch errors?!?).
        y_intersect = shapeIntersection(y_tube, first_invariant{i_input_samples});
      end
        
      % Sampled data viability kernel is the union of invariance kernels
      % (zero sublevel sets) over all input samples.
      if isempty(y_end)
        y_end = y_intersect;
      else
        % Note that this call will generate warnings about empty implicit
        % surface functions for some input values; that is expected.
        y_end = shapeUnion(y_end, y_intersect);
      end
      
    end

    % Reinitialize if necessary.
    if perform_reinit
      start_reinit = cputime;
      y_end = postTimestepReinit(0, y_end, scheme_data_set);
      fprintf('  Finished reinitialization in %g seconds.\n', cputime - start_reinit);
    end
    
    % Get back the correctly shaped data array
    data = reshape(y_end, g.shape);

    % Record result for this step.
    kernels{step_now + 1} = data;
    
    fprintf('  Finished time sample step %d in %g seconds.\n\n', step_now, cputime - start_time_sample);

    if(pause_after_plot)
      % Wait for last plot to be digested.
      pause;
    end

    % Delete last visualization if necessary.
    if(delete_last_plot)
      delete(h);
    end

    % Create new visualization of the viability kernel.
    h = visualizeLevelSet(g, kernels{step_now + 1}, display_type, level, sprintf('constraint set K_{%d}', step_now));
    set(h, 'LineStyle', '-', 'LineWidth', 1, 'LineColor', 'black');
    daspect([ 1 1 1 ]);
    drawnow;
    
    % Increment the number of time samples we have taken.
    step_now = step_now + 1;

  end

  %---------------------------------------------------------------------------
  %% Finished computing the inevitable collision region.
  
  fprintf('\nViability kernel computation time %g seconds\n', cputime - start_time_kernel);
      
  fprintf('\nTotal computation time %g seconds\n', cputime - start_time_overall);

end
